import asyncio
import logging
import os
from unittest import IsolatedAsyncioTestCase as TestCase

import asyncpg
from asyncpg.pool import Pool

from abrechnung.application.users import UserService
from abrechnung.config import Config
from abrechnung.database import revisions
from abrechnung.domain.users import User

lock = asyncio.Lock()


def get_test_db_config() -> dict:
    return {
        "user": os.environ.get("TEST_DB_USER", "abrechnung-test"),
        "password": os.environ.get("TEST_DB_PASSWORD", "asdf1234"),
        "host": os.environ.get("TEST_DB_HOST", "localhost"),
        "dbname": os.environ.get("TEST_DB_DATABASE", "abrechnung-test"),
        "port": int(os.environ.get("TEST_DB_PORT", 5432)),
    }


TEST_CONFIG = {
    "api": {
        "secret_key": "asdf",
        "host": "localhost",
        "port": 8000,
    },
    "registration": {"enabled": True},
    "database": get_test_db_config(),
    "service": {
        "url": "https://abrechnung.example.lol",
        "api_url": "https://abrechnung.example.lol",
        "name": "Test Abrechnung",
    },
}


async def get_test_db() -> Pool:
    """
    get a connection pool to the test database
    """
    cfg = get_test_db_config()
    pool = await asyncpg.create_pool(
        user=cfg["user"],
        password=cfg["password"],
        database=cfg["dbname"],
        host=cfg["host"],
        port=cfg["port"],
        min_size=5,
        max_size=5,
    )

    await revisions.reset_schema(pool)
    await revisions.apply_revisions(pool)

    return pool


class AsyncTestCase(TestCase):
    def __init__(self, *args, log_level=logging.DEBUG, **kwargs):
        super().__init__(*args, **kwargs)
        logging.basicConfig(level=log_level)

    async def _create_test_user(self, username: str, email: str) -> tuple[User, str]:
        """returns the user id and password"""
        # pylint: disable=protected-access
        async with self.db_pool.acquire() as conn:
            password = "asdf1234"
            hashed_password = UserService._hash_password(password)
            user_id = await conn.fetchval(
                "insert into usr (username, email, hashed_password, pending) values ($1, $2, $3, false) returning id",
                username,
                email,
                hashed_password,
            )
            user = await self.user_service.get_user(user_id=user_id)

            return user, password

    def setUp(self) -> None:
        try:
            self.loop = asyncio.get_running_loop()
        except (AttributeError, RuntimeError):  # AttributeError->py36
            self.loop = asyncio.get_event_loop_policy().get_event_loop()

        self.loop.run_until_complete(self._setup_db())
        self.loop.run_until_complete(self.setUpAsync())

    async def _setup_db(self):
        await lock.acquire()

        self.db_pool = await get_test_db()
        self.db_conn: asyncpg.Connection = await self.db_pool.acquire()

    async def _teardown_db(self):
        await self.db_conn.close()
        await self.db_pool.close()

        lock.release()

    async def setUpAsync(self) -> None:
        self.secret_key = "asdf1234"
        self.test_config = Config.from_dict(TEST_CONFIG)
        self.user_service = UserService(self.db_pool, config=self.test_config)

    def tearDown(self) -> None:
        self.loop.run_until_complete(self.tearDownAsync())
        self.loop.run_until_complete(self._teardown_db())

    async def tearDownAsync(self) -> None:
        pass



from aiohttp import web
from marshmallow import Schema, fields

from abrechnung.http.openapi import docs, json_schema
from abrechnung.http.serializers import AccountSchema, SharesField
from abrechnung.http.utils import json_response, PrefixedRouteTableDef

routes = PrefixedRouteTableDef("/api")


async def _account_response(request, account_id: int) -> web.Response:
    account = await request.app["account_service"].get_account(
        user=request["user"], account_id=account_id
    )

    serializer = AccountSchema()
    return json_response(data=serializer.dump(account))


@routes.get(r"/v1/groups/{group_id:\d+}/accounts")
@docs(
    tags=["accounts"],
    summary="list all accounts in a group",
    description="",
    responses={200: {"schema": AccountSchema(many=True), "description": ""}},
)
async def list_accounts(request):
    try:
        accounts = await request.app["account_service"].list_accounts(
            user=request["user"],
            group_id=int(request.match_info["group_id"]),
        )
    except PermissionError:
        raise web.HTTPForbidden(reason="permission denied")

    serializer = AccountSchema()
    return json_response(data=serializer.dump(accounts, many=True))


class AccountUpdateSchema(Schema):
    id = fields.Int()
    type = fields.Str()
    name = fields.Str()
    description = fields.Str()
    priority = fields.Int()
    owning_user_id = fields.Int(allow_none=True)
    clearing_shares = SharesField()
    deleted = fields.Bool()


@routes.post(r"/v1/groups/{group_id:\d+}/accounts/sync")
@docs(
    tags=["accounts"],
    summary="update a collection of accounts",
    description="",
)
@json_schema(AccountUpdateSchema)
async def sync_accounts(request: web.Request):
    updated_ids = await request.app["account_service"].sync_accounts(
        user=request["user"],
        group_id=int(request.match_info["group_id"]),
        accounts=request["json"],  # TODO: FIXME, convert to proper parameters
    )

    await json_response(data=updated_ids)


@routes.post(r"/v1/groups/{group_id:\d+}/accounts")
@docs(
    tags=["accounts"],
    summary="create a new group account",
    description="",
    responses={200: {"schema": AccountSchema(), "description": ""}},
)
@json_schema(
    Schema.from_dict(
        {
            "name": fields.Str(),
            "description": fields.Str(),
            "type": fields.Str(),
            "owning_user_id": fields.Int(
                allow_none=True, load_default=None, required=False
            ),
            "clearing_shares": fields.Dict(
                fields.Int(), fields.Number(), allow_none=True, load_default=None
            ),
        },
        name="CreateAccountSchema",
    )
)
async def create_account(request: web.Request):
    data = request["json"]
    account_id = await request.app["account_service"].create_account(
        user=request["user"],
        group_id=int(request.match_info["group_id"]),
        name=data["name"],
        description=data["description"],
        type=data["type"],
        owning_user_id=data["owning_user_id"],
        clearing_shares=data["clearing_shares"],
    )

    return await _account_response(request=request, account_id=account_id)


@routes.get(r"/v1/accounts/{account_id:\d+}")
@docs(
    tags=["accounts"],
    summary="fetch a group account",
    description="",
    responses={200: {"schema": AccountSchema(), "description": ""}},
)
async def get_account(request: web.Request):
    account_id = int(request.match_info["account_id"])
    return await _account_response(request=request, account_id=account_id)


@routes.post(r"/v1/accounts/{account_id:\d+}")
@docs(
    tags=["accounts"],
    summary="update an account",
    description="",
    responses={200: {"schema": AccountSchema(), "description": ""}},
)
@json_schema(
    Schema.from_dict(
        {
            "name": fields.Str(),
            "description": fields.Str(),
            "owning_user_id": fields.Int(
                allow_none=True, load_default=None, required=False
            ),
            "clearing_shares": fields.Dict(
                fields.Int(), fields.Number(), allow_none=True, load_default=None
            ),
        },
        name="UpdateAccountSchema",
    )
)
async def update_account(request: web.Request):
    account_id = int(request.match_info["account_id"])
    data = request["json"]
    await request.app["account_service"].update_account(
        user=request["user"],
        account_id=account_id,
        name=data["name"],
        description=data["description"],
        owning_user_id=data["owning_user_id"],
        clearing_shares=data["clearing_shares"],
    )

    return await _account_response(request=request, account_id=account_id)


@routes.delete(r"/v1/accounts/{account_id:\d+}")
@docs(
    tags=["accounts"],
    summary="delete an account",
    description="",
    responses={200: {"schema": AccountSchema(), "description": ""}},
)
async def delete_account(request: web.Request):
    account_id = int(request.match_info["account_id"])
    await request.app["account_service"].delete_account(
        user=request["user"],
        account_id=account_id,
    )

    return await _account_response(request=request, account_id=account_id)



import json
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Optional, Union

import asyncpg

from abrechnung.domain.accounts import Account, AccountType, AccountDetails
from abrechnung.util import parse_postgres_datetime
from . import (
    Application,
    NotFoundError,
    check_group_permissions,
    InvalidCommand,
    create_group_log,
)
from ..domain.users import User


@dataclass
class RawAccount:
    id: int
    group_id: int
    name: str
    description: str
    priority: int
    owning_user_id: Optional[int]
    deleted: bool
    type: Optional[str] = field(default=None)
    clearing_shares: Optional[dict[int, float]] = field(default=None)


class AccountService(Application):
    @staticmethod
    async def _increment_account_version(conn: asyncpg.Connection, revision_id: int):
        await conn.execute(
            "update account_revision set version = version + 1 where id = $1",
            revision_id,
        )

    @staticmethod
    async def _get_or_create_revision(
        conn: asyncpg.Connection, user: User, account_id: int
    ) -> int:
        """return the revision id, assumes we are already in a transaction"""
        revision_id = await conn.fetchval(
            "select id "
            "from account_revision "
            "where account_id = $1 and user_id = $2 and committed is null",
            account_id,
            user.id,
        )
        if revision_id:  # there already is a wip revision
            return revision_id

        # create a new transaction revision
        revision_id = await conn.fetchval(
            "insert into account_revision (user_id, account_id) values ($1, $2) returning id",
            user.id,
            account_id,
        )
        return revision_id

    @staticmethod
    async def _check_account_permissions(
        conn: asyncpg.Connection,
        user: User,
        account_id: int,
        can_write: bool = False,
        account_type: Optional[Union[str, list[str]]] = None,
    ) -> tuple[int, str]:
        """returns group id of the transaction"""
        result = await conn.fetchrow(
            "select a.type, a.group_id, can_write, is_owner "
            "from group_membership gm join account a on gm.group_id = a.group_id and gm.user_id = $1 "
            "where a.id = $2",
            user.id,
            account_id,
        )
        if not result:
            raise NotFoundError(f"account not found")

        if can_write and not (result["can_write"] or result["is_owner"]):
            raise PermissionError(f"user does not have write permissions")

        if account_type:
            type_check = (
                [account_type] if isinstance(account_type, str) else account_type
            )
            if result["type"] not in type_check:
                raise InvalidCommand(
                    f"Transaction type {result['type']} does not match the expected type {type_check}"
                )

        return result["group_id"], result["type"]

    async def _get_or_create_pending_account_change(
        self, conn: asyncpg.Connection, user: User, account_id: int
    ) -> int:
        revision_id = await self._get_or_create_revision(
            conn=conn, user=user, account_id=account_id
        )

        a = await conn.fetchval(
            "select id from account_history th where revision_id = $1 and id = $2",
            revision_id,
            account_id,
        )
        if a:
            return revision_id

        last_committed_revision = await conn.fetchval(
            "select ar.id "
            "from account_revision ar "
            "   join account_history ah on ar.id = ah.revision_id and ar.account_id = ah.id "
            "where ar.account_id = $1 and ar.committed is not null "
            "order by ar.committed desc "
            "limit 1",
            account_id,
        )

        if last_committed_revision is None:
            raise InvalidCommand(
                f"Cannot edit account {account_id} as it has no committed changes."
            )

        # copy all existing transaction data into a new history entry
        await conn.execute(
            "insert into account_history (id, revision_id, name, description, priority, deleted)"
            "select id, $1, name, description, priority, deleted "
            "from account_history where id = $2 and revision_id = $3",
            revision_id,
            account_id,
            last_committed_revision,
        )

        # copy all last committed creditor shares
        await conn.execute(
            "insert into clearing_account_share (account_id, revision_id, share_account_id, shares) "
            "select account_id, $1, share_account_id, shares "
            "from clearing_account_share where account_id = $2 and revision_id = $3",
            revision_id,
            account_id,
            last_committed_revision,
        )

        return revision_id

    @staticmethod
    async def _check_account_exists(
        conn: asyncpg.Connection, group_id: int, account_id: int
    ) -> int:
        acc = await conn.fetchval(
            "select account_id from committed_account_state_valid_at() "
            "where group_id = $1 and account_id = $2 and not deleted",
            group_id,
            account_id,
        )
        if not acc:
            raise NotFoundError(f"Account with id {account_id}")

        return True

    async def _account_clearing_shares_check(
        self,
        conn: asyncpg.Connection,
        user: User,
        account_id: int,
        share_account_id: int,
        account_type: Optional[Union[str, list[str]]] = None,
    ) -> tuple[int, int]:
        """returns tuple of group_id of the account and the users revision_id of the pending change"""
        group_id, _ = await self._check_account_permissions(
            conn=conn,
            user=user,
            account_id=account_id,
            can_write=True,
            account_type=account_type,
        )

        await self._check_account_exists(conn, group_id, share_account_id)

        revision_id = await self._get_or_create_pending_account_change(
            conn=conn, user=user, account_id=account_id
        )

        return group_id, revision_id

    @staticmethod
    def _account_detail_from_db_json(db_json: dict) -> AccountDetails:
        return AccountDetails(
            name=db_json["name"],
            description=db_json["description"],
            deleted=db_json["deleted"],
            priority=db_json["priority"],
            owning_user_id=db_json["owning_user_id"],
            revision_started_at=parse_postgres_datetime(db_json["revision_started"]),
            revision_committed_at=None
            if db_json.get("revision_committed") is None
            else parse_postgres_datetime(db_json["revision_committed"]),
            clearing_shares={
                cred["share_account_id"]: cred["shares"]
                for cred in db_json["clearing_shares"]
            },
            changed_by=db_json["changed_by"],
        )

    def _account_db_row(self, account: asyncpg.Record) -> Account:
        committed_details = (
            self._account_detail_from_db_json(
                json.loads(account["committed_details"])[0]
            )
            if account["committed_details"]
            else None
        )
        pending_details = (
            self._account_detail_from_db_json(json.loads(account["pending_details"])[0])
            if account["pending_details"]
            else None
        )

        return Account(
            id=account["account_id"],
            group_id=account["group_id"],
            type=account["type"],
            is_wip=account["is_wip"],
            last_changed=account["last_changed"],
            version=account["version"],
            committed_details=committed_details,
            pending_details=pending_details,
        )

    async def list_accounts(self, *, user: User, group_id: int) -> list[Account]:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                await check_group_permissions(conn=conn, group_id=group_id, user=user)
                cur = conn.cursor(
                    "select account_id, group_id, type, last_changed, version, is_wip, "
                    "   committed_details, pending_details "
                    "from full_account_state_valid_at($1) "
                    "where group_id = $2",
                    user.id,
                    group_id,
                )

                result = []
                async for account in cur:
                    result.append(self._account_db_row(account))

                return result

    async def get_account(self, *, user: User, account_id: int) -> Account:
        async with self.db_pool.acquire() as conn:
            await self._check_account_permissions(
                conn=conn, user=user, account_id=account_id
            )
            account = await conn.fetchrow(
                "select account_id, group_id, type, last_changed, version, is_wip, "
                "   committed_details, pending_details "
                "from full_account_state_valid_at($1) "
                "where account_id = $2",
                user.id,
                account_id,
            )
            return self._account_db_row(account)

    async def _create_account(
        self,
        *,
        conn: asyncpg.Connection,
        user: User,
        group_id: int,
        type: str,
        name: str,
        description: str,
        owning_user_id: Optional[int] = None,
        priority: int = 0,
        clearing_shares: Optional[dict[int, float]] = None,
    ) -> int:
        if clearing_shares and type != AccountType.clearing.value:
            raise InvalidCommand(
                f"'{type}' accounts cannot have associated settlement distribution shares"
            )

        can_write, is_owner = await check_group_permissions(
            conn=conn, group_id=group_id, user=user, can_write=True
        )
        if owning_user_id is not None:
            if not is_owner and owning_user_id != user.id:
                raise PermissionError(
                    f"only group owners can associate others with accounts"
                )

        account_id = await conn.fetchval(
            "insert into account (group_id, type) values ($1, $2) returning id",
            group_id,
            type,
        )

        revision_id = await self._get_or_create_revision(conn, user, account_id)
        await conn.execute(
            "insert into account_history (id, revision_id, name, description, owning_user_id, priority) "
            "values ($1, $2, $3, $4, $5, $6)",
            account_id,
            revision_id,
            name,
            description,
            owning_user_id,
            priority,
        )
        if clearing_shares and type == AccountType.clearing.value:
            # TODO: make this more efficient
            for share_account_id, value in clearing_shares.items():
                if value == 0:
                    continue
                await self._check_account_exists(conn, group_id, share_account_id)
                await conn.execute(
                    "insert into clearing_account_share (account_id, revision_id, share_account_id, shares) "
                    "values ($1, $2, $3, $4)",
                    account_id,
                    revision_id,
                    share_account_id,
                    value,
                )

        await create_group_log(
            conn=conn,
            group_id=group_id,
            user=user,
            type="account-committed",
            message=f"created account {name}",
        )
        await conn.execute(
            "update account_revision set committed = now() where id = $1",
            revision_id,
        )
        return account_id

    async def create_account(
        self,
        *,
        user: User,
        group_id: int,
        type: str,
        name: str,
        description: str,
        owning_user_id: Optional[int] = None,
        priority: int = 0,
        clearing_shares: Optional[dict[int, float]] = None,
    ) -> int:
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                return await self._create_account(
                    conn=conn,
                    user=user,
                    group_id=group_id,
                    type=type,
                    name=name,
                    description=description,
                    owning_user_id=owning_user_id,
                    priority=priority,
                    clearing_shares=clearing_shares,
                )

    async def _update_account(
        self,
        conn: asyncpg.Connection,
        user: User,
        account_id: int,
        name: str,
        description: str,
        owning_user_id: Optional[int] = None,
        priority: int = 0,
        clearing_shares: Optional[dict[int, float]] = None,
    ):
        group_id, account_type = await self._check_account_permissions(
            conn=conn, user=user, account_id=account_id, can_write=True
        )
        can_write, is_owner = await check_group_permissions(
            conn=conn, group_id=group_id, user=user, can_write=True
        )
        if clearing_shares and account_type != AccountType.clearing.value:
            raise InvalidCommand(
                f"'{account_type}' accounts cannot have associated settlement distribution shares"
            )

        revision_id = await self._get_or_create_revision(
            conn=conn, user=user, account_id=account_id
        )

        committed_account = await conn.fetchrow(
            "select owning_user_id from committed_account_state_valid_at() where account_id = $1",
            account_id,
        )

        if owning_user_id is not None:
            if not is_owner and owning_user_id != user.id:
                raise PermissionError(
                    f"only group owners can associate others with accounts"
                )
        elif (
            committed_account["owning_user_id"] is not None
            and committed_account["owning_user_id"] != user.id
            and not is_owner
        ):
            raise PermissionError(
                f"only group owners can remove other users as account owners"
            )

        await conn.execute(
            "insert into account_history (id, revision_id, name, description, owning_user_id, priority) "
            "values ($1, $2, $3, $4, $5, $6) ",
            account_id,
            revision_id,
            name,
            description,
            owning_user_id,
            priority,
        )
        if clearing_shares and account_type == AccountType.clearing.value:
            # TODO: make this more efficient
            for share_account_id, value in clearing_shares.items():
                if value == 0:
                    continue
                await self._check_account_exists(conn, group_id, share_account_id)
                await conn.execute(
                    "insert into clearing_account_share (account_id, revision_id, share_account_id, shares) "
                    "values ($1, $2, $3, $4)",
                    account_id,
                    revision_id,
                    share_account_id,
                    value,
                )

        await self._increment_account_version(conn=conn, revision_id=revision_id)
        await create_group_log(
            conn=conn,
            group_id=group_id,
            user=user,
            type="account-committed",
            message=f"updated account {name}",
        )
        await conn.execute(
            "update account_revision set committed = now() where id = $1",
            revision_id,
        )

    async def update_account(
        self,
        user: User,
        account_id: int,
        name: str,
        description: str,
        owning_user_id: Optional[int] = None,
        priority: int = 0,
        clearing_shares: Optional[dict[int, float]] = None,
    ):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                return await self._update_account(
                    conn=conn,
                    user=user,
                    account_id=account_id,
                    name=name,
                    description=description,
                    owning_user_id=owning_user_id,
                    priority=priority,
                    clearing_shares=clearing_shares,
                )

    async def delete_account(
        self,
        user: User,
        account_id: int,
    ):
        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                group_id, _ = await self._check_account_permissions(
                    conn=conn, user=user, account_id=account_id, can_write=True
                )
                row = await conn.fetchrow(
                    "select id from account where id = $1",
                    account_id,
                )
                if row is None:
                    raise InvalidCommand(f"Account does not exist")

                has_committed_shares = await conn.fetchval(
                    "select 1 "
                    "from committed_transaction_state_valid_at() t "
                    "where not deleted and $1 = any(involved_accounts)",
                    account_id,
                )
                has_pending_shares = await conn.fetchval(
                    "select 1 "
                    "from aggregated_pending_transaction_history t "
                    "where $1 = any(t.involved_accounts)",
                    account_id,
                )

                has_committed_usages = await conn.fetchval(
                    "select 1 "
                    "from committed_transaction_position_state_valid_at() p "
                    "join transaction t on t.id = p.transaction_id "
                    "where not p.deleted and $1 = any(p.involved_accounts)",
                    account_id,
                )

                has_pending_usages = await conn.fetchval(
                    "select 1 "
                    "from aggregated_pending_transaction_position_history p "
                    "join transaction t on t.id = p.transaction_id "
                    "where $1 = any(p.involved_accounts)",
                    account_id,
                )

                if (
                    has_committed_shares
                    or has_pending_shares
                    or has_committed_usages
                    or has_pending_usages
                ):
                    raise InvalidCommand(
                        f"Cannot delete an account that is references by a transaction"
                    )

                row = await conn.fetchrow(
                    "select name, revision_id, deleted "
                    "from committed_account_state_valid_at() "
                    "where account_id = $1",
                    account_id,
                )
                if row is None:
                    raise InvalidCommand(
                        f"Cannot delete an account without any committed changes"
                    )

                if row["deleted"]:
                    raise InvalidCommand(f"Cannot delete an already deleted account")

                has_committed_clearing_shares = await conn.fetchval(
                    "select 1 "
                    "from committed_account_state_valid_at() p "
                    "where not p.deleted and $1 = any(p.involved_accounts)",
                    account_id,
                )

                has_pending_clearing_shares = await conn.fetchval(
                    "select 1 "
                    "from aggregated_pending_account_history p "
                    "where $1 = any(p.involved_accounts)",
                    account_id,
                )
                if has_committed_clearing_shares or has_pending_clearing_shares:
                    raise InvalidCommand(
                        f"Cannot delete an account that is references by another clearing account"
                    )

                now = datetime.now(tz=timezone.utc)
                revision_id = await conn.fetchval(
                    "insert into account_revision (user_id, account_id, started, committed) "
                    "values ($1, $2, $3, $4) returning id",
                    user.id,
                    account_id,
                    now,
                    now,
                )
                await conn.execute(
                    "insert into account_history (id, revision_id, name, description, priority, deleted) "
                    "select $1, $2, name, description, priority, true "
                    "from account_history ah where ah.id = $1 and ah.revision_id = $3 ",
                    account_id,
                    revision_id,
                    row["revision_id"],
                )

                await create_group_log(
                    conn=conn,
                    group_id=group_id,
                    user=user,
                    type="account-deleted",
                    message=f"deleted account account {row['name']}",
                )

    async def sync_account(
        self, *, conn: asyncpg.Connection, user: User, account: RawAccount
    ) -> tuple[int, int]:
        if account.id > 0:
            await self._update_account(
                conn=conn,
                user=user,
                account_id=account.id,
                name=account.name,
                description=account.description,
                owning_user_id=account.owning_user_id,
                priority=account.priority,
                clearing_shares=account.clearing_shares,
            )
            return account.id, account.id

        if account.type is None:
            raise InvalidCommand("new account must have 'type' set")

        new_acc_id = await self._create_account(
            conn=conn,
            user=user,
            group_id=account.group_id,
            type=account.type,
            name=account.name,
            description=account.description,
            owning_user_id=account.owning_user_id,
            priority=account.priority,
            clearing_shares=account.clearing_shares,
        )
        return account.id, new_acc_id

    async def sync_accounts(
        self, *, user: User, group_id: int, accounts: list[RawAccount]
    ) -> dict[int, int]:
        all_accounts_in_same_group = all([a.group_id == group_id for a in accounts])
        if not all_accounts_in_same_group:
            raise InvalidCommand("all accounts must belong to the same group")

        async with self.db_pool.acquire() as conn:
            async with conn.transaction():
                can_write, _ = await check_group_permissions(
                    conn=conn, group_id=group_id, user=user, can_write=True
                )

                if not can_write:
                    raise PermissionError("need write access to group")

                new_account_id_map: dict[int, int] = {}

                for account in filter(
                    lambda acc: acc.type == AccountType.personal, accounts
                ):
                    old_acc_id, new_acc_id = await self.sync_account(
                        conn=conn, user=user, account=account
                    )
                    new_account_id_map[old_acc_id] = new_acc_id

                clearing_accounts = list(
                    filter(lambda acc: acc.type == AccountType.clearing, accounts)
                )
                # TODO: improve this very inefficient implementation
                # first step: use a dict instead of a list
                while len(clearing_accounts) > 0:
                    for account in clearing_accounts[:]:  # copy as we remove items
                        if account.clearing_shares:
                            account.clearing_shares = {
                                new_account_id_map.get(k, k): v
                                for k, v in account.clearing_shares.items()
                            }
                        if account.clearing_shares and all(
                            [x > 0 for x in account.clearing_shares.keys()]
                        ):
                            old_acc_id, new_acc_id = await self.sync_account(
                                conn=conn, user=user, account=account
                            )
                            new_account_id_map[old_acc_id] = new_acc_id
                            clearing_accounts.remove(account)

                return new_account_id_map



import argparse
import logging
from datetime import datetime

from abrechnung import subcommand
from abrechnung.config import Config
from abrechnung.database import db_connect


class Demo(subcommand.SubCommand):
    def __init__(self, config: Config, **args):  # pylint: disable=super-init-not-called
        self.config = config
        self.logger = logging.getLogger(__name__)

        self.command = args.pop("command")
        self.args = args

    @staticmethod
    def argparse_register(subparser: argparse.ArgumentParser):
        subparsers = subparser.add_subparsers(required=True)

        create_user_parser = subparsers.add_parser("cleanup")
        create_user_parser.set_defaults(command="cleanup")

    async def handle_cleanup_command(self):
        if not self.config["demo"]["enabled"]:
            self.logger.error("This instance is not running in demo mode, exiting")
            return

        deletion_threshold = datetime.now() - self.config["demo"]["wipe_interval"]

        db_pool = await db_connect(self.config["database"])
        async with db_pool.acquire() as conn:
            async with conn.transaction():
                n_rows_groups = await conn.fetchval(
                    "delete from grp where id in ( "
                    "   select g.id from grp g "
                    "   join group_membership gm on gm.group_id = grp.id "
                    "   join usr on usr.id = gm.user_id "
                    "   where registered_at <= $1"
                    ") returning *",
                    deletion_threshold,
                )

                n_rows_users = await conn.fetchval(
                    "delete from usr where registered_at <= $1 returning *",
                    deletion_threshold,
                )

                n_deleted = (n_rows_groups or 0) + (n_rows_users or 0)
                self.logger.info(
                    f"Deleted demo users and groups, a total of {n_deleted} database rows"
                )

    async def run(self):
        if self.command == "cleanup":
            return await self.handle_cleanup_command()



from datetime import datetime

from aiohttp import web
from aiohttp.abc import Request
from aiohttp.web_request import FileField
from marshmallow import Schema, fields

from abrechnung.http.openapi import docs, json_schema
from abrechnung.http.serializers import (
    TransactionSchema,
    TransactionPositionSchema,
    SharesField,
)
from abrechnung.http.utils import json_response, PrefixedRouteTableDef

routes = PrefixedRouteTableDef("/api")


class TransactionUpdateScheme(Schema):
    description = fields.Str()
    value = fields.Number()
    currency_symbol = fields.Str()
    currency_conversion_rate = fields.Number()
    billed_at = fields.Date()
    revision_started_at = fields.DateTime()
    revision_committed_at = fields.DateTime(required=False)
    creditor_shares = SharesField()
    debitor_shares = SharesField()
    deleted = fields.Bool()
    positions = fields.List(fields.Nested(TransactionPositionSchema), dump_default=[])


async def _transaction_response(request, transaction_id: int) -> web.Response:
    transaction = await request.app["transaction_service"].get_transaction(
        user=request["user"], transaction_id=transaction_id
    )

    serializer = TransactionSchema()
    return json_response(data=serializer.dump(transaction))


@routes.get(r"/v1/groups/{group_id:\d+}/transactions")
@docs(
    tags=["transactions"],
    summary="list all transactions in a group",
    responses={200: {"schema": TransactionSchema(many=True), "description": ""}},
    description="",
)
async def list_transactions(request):
    group_id: int = int(request.match_info["group_id"])
    min_last_changed = request.query.get("min_last_changed")

    if min_last_changed:
        try:
            min_last_changed = datetime.fromisoformat(min_last_changed)
        except ValueError:
            raise web.HTTPBadRequest(
                reason="Invalid query param 'min_last_changed', must be a valid ISO timestamp."
            )

    forced_transaction_ids = request.query.get("transaction_ids")
    if forced_transaction_ids:
        try:
            forced_transaction_ids = [int(x) for x in forced_transaction_ids.split(",")]
        except ValueError:
            raise web.HTTPBadRequest(
                reason="Invalid query param 'transaction_ids', must be a comma separated list of integers"
            )

    transactions = await request.app["transaction_service"].list_transactions(
        user=request["user"],
        group_id=group_id,
        min_last_changed=min_last_changed,
        additional_transactions=forced_transaction_ids,
    )

    serializer = TransactionSchema()
    return json_response(data=serializer.dump(transactions, many=True))


@routes.post(r"/v1/groups/{group_id:\d+}/transactions/sync")
@docs(tags=["transactions"], summary="sync a batch of transactions", description="")
@json_schema(TransactionUpdateScheme(many=True))
async def sync_transactions(request: Request):
    updated_ids = await request.app["transaction_service"].sync_transactions(
        user=request["user"],
        group_id=int(request.match_info["group_id"]),
        transactions=request["json"],  # TODO: FIXME, convert to proper parameters
    )

    await json_response(data=updated_ids)


@routes.post(r"/v1/groups/{group_id:\d+}/transactions")
@docs(
    tags=["transactions"],
    summary="create a new transaction",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
@json_schema(
    Schema.from_dict(
        {
            "description": fields.Str(),
            "type": fields.Str(),
            "value": fields.Number(),
            "currency_symbol": fields.Str(),
            "billed_at": fields.Date(),
            "currency_conversion_rate": fields.Number(),
            "creditor_shares": fields.Dict(
                fields.Int(), fields.Number(), load_default=None, required=False
            ),
            "debitor_shares": fields.Dict(
                fields.Int(), fields.Number(), load_default=None, required=False
            ),
            "positions": fields.List(
                fields.Nested(TransactionPositionSchema),
                required=False,
                nullable=True,
                load_default=None,
            ),
            "perform_commit": fields.Bool(load_default=False, required=False),
        },
        name="CreateTransactionSchema",
    )
)
async def create_transaction(request: Request):
    data = request["json"]
    group_id: int = int(request.match_info["group_id"])

    transaction_id = await request.app["transaction_service"].create_transaction(
        user=request["user"],
        group_id=group_id,
        description=data["description"],
        type=data["type"],
        currency_symbol=data["currency_symbol"],
        currency_conversion_rate=float(data["currency_conversion_rate"]),
        billed_at=data["billed_at"],
        value=float(data["value"]),
        creditor_shares=data["creditor_shares"],
        debitor_shares=data["debitor_shares"],
        positions=data["positions"],
        perform_commit=data["perform_commit"],
    )

    return await _transaction_response(request, transaction_id)


@routes.get(r"/v1/transactions/{transaction_id:\d+}")
@docs(
    tags=["transactions"],
    summary="get transaction details",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
async def get_transaction(request: Request):
    transaction = await request.app["transaction_service"].get_transaction(
        user=request["user"],
        transaction_id=int(request.match_info["transaction_id"]),
    )

    serializer = TransactionSchema()
    return json_response(data=serializer.dump(transaction))


@routes.post(r"/v1/transactions/{transaction_id:\d+}")
@docs(
    tags=["transactions"],
    summary="update transaction details",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
@json_schema(
    Schema.from_dict(
        {
            "description": fields.Str(),
            "value": fields.Number(),
            "currency_symbol": fields.Str(),
            "billed_at": fields.Date(),
            "currency_conversion_rate": fields.Number(),
            "creditor_shares": fields.Dict(
                fields.Int(), fields.Number(), load_default=None, required=False
            ),
            "debitor_shares": fields.Dict(
                fields.Int(), fields.Number(), load_default=None, required=False
            ),
            "positions": fields.List(
                fields.Nested(TransactionPositionSchema),
                required=False,
                nullable=True,
                load_default=None,
            ),
            "perform_commit": fields.Bool(load_default=False, required=False),
        },
        name="UpdateTransactionSchema",
    )
)
async def update_transaction(request: Request):
    data = request["json"]
    transaction_id = int(request.match_info["transaction_id"])

    await request.app["transaction_service"].update_transaction(
        user=request["user"],
        transaction_id=transaction_id,
        value=data["value"],
        description=data["description"],
        currency_symbol=data["currency_symbol"],
        currency_conversion_rate=data["currency_conversion_rate"],
        billed_at=data["billed_at"],
        creditor_shares=data["creditor_shares"],
        debitor_shares=data["debitor_shares"],
        positions=data["positions"],
        perform_commit=data["perform_commit"],
    )

    return await _transaction_response(request, transaction_id)


@routes.post(r"/v1/transactions/{transaction_id:\d+}/positions")
@docs(
    tags=["transactions"],
    summary="update transaction positions",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
@json_schema(
    Schema.from_dict(
        {
            "positions": fields.List(
                fields.Nested(TransactionPositionSchema),
            ),
            "perform_commit": fields.Bool(load_default=False, required=False),
        },
        name="UpdateTransactionSchemaV2",
    )
)
async def update_transaction_positions(request: Request):
    data = request["json"]
    transaction_id = int(request.match_info["transaction_id"])

    await request.app["transaction_service"].update_transaction_positions(
        user=request["user"],
        transaction_id=transaction_id,
        positions=data["positions"],
        perform_commit=data["perform_commit"],
    )

    return await _transaction_response(request, transaction_id)


@routes.post(r"/v1/transactions/{transaction_id:\d+}/commit")
@docs(
    tags=["transactions"],
    summary="commit currently pending transaction changes",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
async def commit_transaction(request: Request):
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].commit_transaction(
        user=request["user"],
        transaction_id=transaction_id,
    )

    return await _transaction_response(request, transaction_id)


@routes.delete(r"/v1/transactions/{transaction_id:\d+}")
@docs(
    tags=["transactions"],
    summary="delete a transaction",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
async def delete_transaction(request: Request):
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].delete_transaction(
        user=request["user"],
        transaction_id=transaction_id,
    )

    return await _transaction_response(request, transaction_id)


@routes.post(r"/v1/transactions/{transaction_id:\d+}/new_change")
@docs(
    tags=["transactions"],
    summary="create a new pending transaction revision",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
async def create_transaction_change(request: Request):
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].create_transaction_change(
        user=request["user"],
        transaction_id=int(request.match_info["transaction_id"]),
    )

    return await _transaction_response(request, transaction_id)


@routes.post(r"/v1/transactions/{transaction_id:\d+}/discard")
@docs(
    tags=["transactions"],
    summary="discard currently pending transaction changes",
    responses={200: {"schema": TransactionSchema, "description": ""}},
    description="",
)
async def discard_transaction_change(request: Request):
    transaction_id = int(request.match_info["transaction_id"])
    await request.app["transaction_service"].discard_transaction_changes(
        user=request["user"],
        transaction_id=transaction_id,
    )

    return await _transaction_response(request, transaction_id)


@routes.post(r"/v1/transactions/{transaction_id:\d+}/files")
@docs(
    tags=["transactions"],
    summary="upload a file as a transaction attachment",
    description="",
    responses={200: {"schema": TransactionSchema, "description": ""}},
)
async def upload_file(request: Request):
    data = await request.post()

    if "file" not in data or not isinstance(data["file"], FileField):
        raise web.HTTPBadRequest(reason=f"File data required")
    if "filename" not in data or not isinstance(data["filename"], str):
        raise web.HTTPBadRequest(reason=f"File Name required")

    transaction_id = int(request.match_info["transaction_id"])
    file_field: FileField = data["file"]
    # filename contains the name of the file in string format.
    filename = data["filename"]
    input_file = file_field.file
    mime_type = file_field.content_type
    try:
        content = input_file.read()
    except Exception as e:
        raise web.HTTPBadRequest(reason=f"Cannot read uploaded file: {e}")

    await request.app["transaction_service"].upload_file(
        user=request["user"],
        transaction_id=transaction_id,
        filename=filename,
        mime_type=mime_type,
        content=content,
    )

    return await _transaction_response(request, transaction_id)


@routes.delete(r"/v1/files/{file_id:\d+}")
@docs(
    tags=["transactions"],
    summary="delete a transaction attachment",
    description="",
    responses={200: {"schema": TransactionSchema, "description": ""}},
)
async def delete_file(request: Request):
    transaction_id, revision_id = await request.app["transaction_service"].delete_file(
        user=request["user"],
        file_id=int(request.match_info["file_id"]),
    )

    return await _transaction_response(request, transaction_id)


@routes.get(r"/v1/files/{file_id:\d+}/{blob_id:\d+}")
@docs(
    tags=["transactions"],
    summary="fetch the (binary) contents of a transaction attachment",
    description="",
)
async def get_file_contents(request: Request):
    file_id = int(request.match_info["file_id"])
    blob_id = int(request.match_info["blob_id"])
    mime_type, content = await request.app["transaction_service"].read_file_contents(
        user=request["user"],
        file_id=file_id,
        blob_id=blob_id,
    )

    return web.Response(body=content, content_type=mime_type)



from copy import deepcopy
from datetime import datetime, timedelta

from abrechnung.application.users import UserService

from abrechnung.application.groups import GroupService
from abrechnung.config import Config

from tests import AsyncTestCase, TEST_CONFIG


class TransactionLogicTest(AsyncTestCase):
    async def setUpAsync(self) -> None:
        await super().setUpAsync()
        self.group_service = GroupService(self.db_pool, config=self.test_config)

        self.user, _ = await self._create_test_user("test", "test@test.test")
        self.group_id = await self.group_service.create_group(
            user=self.user,
            name="test group",
            description="",
            currency_symbol="€",
            terms="",
            add_user_account_on_join=False,
        )

    async def test_register_guest_user(self):
        config = deepcopy(TEST_CONFIG)
        config["registration"]["allow_guest_users"] = True
        config["registration"]["valid_email_domains"] = ["stusta.de"]
        test_config = Config.from_dict(config)
        user_service = UserService(self.db_pool, config=test_config)

        invite_token_id = await self.group_service.create_invite(
            user=self.user,
            group_id=self.group_id,
            description="foo",
            single_use=False,
            join_as_editor=False,
            valid_until=datetime.now() + timedelta(hours=1),
        )
        self.assertIsNotNone(invite_token_id)
        invites = await self.group_service.list_invites(
            user=self.user, group_id=self.group_id
        )
        self.assertEqual(1, len(invites))
        invite_token = invites[0].token

        user_id = await user_service.register_user(
            username="guest user 1",
            email="foobar@something.com",
            password="asdf1234",
            invite_token=invite_token,
        )
        self.assertIsNotNone(user_id)
        guest_user = await user_service.get_user(user_id=user_id)
        self.assertTrue(guest_user.is_guest_user)

        user_id = await user_service.register_user(
            username="no guest user 1",
            email="foobar@stusta.de",
            password="asdf1234",
            invite_token=invite_token,
        )
        self.assertIsNotNone(user_id)
        non_guest_user = await user_service.get_user(user_id=user_id)
        self.assertFalse(non_guest_user.is_guest_user)

        with self.assertRaises(PermissionError):
            await user_service.register_user(
                username="invalid user",
                email="foobar@something.com",
                password="asdf1234",
            )



#!/usr/bin/env python
from yaml import safe_load, dump
import sys
from os import execvp, execlp, getenv, makedirs
import subprocess


def to_bool(data: str):
    return data.lower() in [
        "true",
        "1",
        "t",
        "y",
        "yes",
        "on",
    ]


print("generating config")
config = dict()
filename = "/usr/share/abrechnung/config/abrechnung.yaml"
with open(filename, "r", encoding="utf-8") as filehandle:
    config = safe_load(filehandle)

if not "service" in config:
    config["service"] = dict()
if not "database" in config:
    config["database"] = dict()
if not "registration" in config:
    config["registration"] = dict()
if not "email" in config:
    config["email"] = dict()

config["service"]["url"] = getenv("SERVICE_URL", "https://localhost")
config["service"]["api_url"] = getenv("SERVICE_API_URL", "https://localhost/api")
config["service"]["name"] = getenv("SERVICE_NAME", "Abrechnung")
config["database"]["host"] = getenv("DB_HOST")
config["database"]["user"] = getenv("DB_USER")
config["database"]["dbname"] = getenv("DB_NAME")
config["database"]["password"] = getenv("DB_PASSWORD")

config["api"]["secret_key"] = getenv("ABRECHNUNG_SECRET")
config["api"]["host"] = "0.0.0.0"
config["api"]["port"] = int(getenv("ABRECHNUNG_PORT", "8080"))
config["api"]["id"] = getenv("ABRECHNUNG_ID", "default")

config["registration"]["allow_guest_users"] = to_bool(
    getenv("REGISTRATION_ALLOW_GUEST_USERS", "false")
)
config["registration"]["enabled"] = to_bool(getenv("REGISTRATION_ENABLED", "false"))
config["registration"]["valid_email_domains"] = getenv(
    "REGISTRATION_VALID_EMAIL_DOMAINS", "false"
).split(",")

config["email"]["address"] = getenv("ABRECHNUNG_EMAIL", "")
config["email"]["host"] = getenv("SMTP_HOST", "localhost")
config["email"]["port"] = int(getenv("SMTP_PORT", "587"))
config["email"]["mode"] = getenv("SMTP_MODE", "smtp-starttls")

if getenv("SMTP_USER", None):
    if not "auth" in config["email"]:
        config["email"]["auth"] = dict()

    config["email"]["auth"]["username"] = getenv("SMTP_USER", None)
    config["email"]["auth"]["password"] = getenv("SMTP_PASSWORD", None)

output = dump(config)
makedirs("/etc/abrechnung/", exist_ok=True)
with open("/etc/abrechnung/abrechnung.yaml", "w", encoding="utf-8") as file:
    file.write(output)
print("config done")
sys.stdout.flush()
if sys.argv[1] == "api":
    print("migrating")
    sys.stdout.flush()
    subprocess.run("abrechnung db migrate", shell=True, check=True)
    print("migrated")
if sys.argv[1] == "cron":
    print("running cron...")
    sys.stdout.flush()
    execlp("crond", "crond", "-f")
print("starting abrechnung...")
sys.stdout.flush()
execvp("abrechnung", sys.argv)



from typing import Mapping, Any, Optional

from marshmallow import fields, Schema


class SharesField(fields.Field):
    def _serialize(
        self, value: Mapping[int, float], attr: Optional[str], obj: Any, **kwargs
    ):
        return {
            str(account_id): usage_value for account_id, usage_value in value.items()
        }

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ):
        if not isinstance(value, dict):
            raise fields.ValidationError(message="expected a dictionary")
        try:
            return {int(k): float(v) for k, v in value.items()}
        except:
            raise fields.ValidationError(
                message="expected a dictionary mapping int to float"
            )


class TransactionPositionSchema(Schema):
    id = fields.Int()
    price = fields.Number()
    communist_shares = fields.Number()
    deleted = fields.Bool(required=False, load_default=False)
    name = fields.Str()
    usages = SharesField(load_default={})


s = Schema.from_dict(
    {
        "description": fields.Str(),
        "type": fields.Str(),
        "value": fields.Number(),
        "currency_symbol": fields.Str(),
        "billed_at": fields.Date(),
    },
    name="CreateTransactionSchema",
)()

data = {
    "description": "asdfasdf",
    "value": 11,
    "type": "purchase",
    "billed_at": "2022-10-21T18:57:12.628111",
    "currency_symbol": "€",
    "currency_conversion_rate": 1,
    "creditor_shares": {"2": 1},
    "debitor_shares": None,
    "perform_commit": False,
}

print(s.validate(data=data))



import re
from datetime import timedelta
from pathlib import Path
from typing import Any, Optional, Mapping

import yaml
from marshmallow import Schema, fields

config_timedelta_format = re.compile(
    r"((?P<hours>\d+?)h)?((?P<minutes>\d+?)m)?((?P<seconds>\d+?)s)?"
)


class TimedeltaField(fields.Field):
    def _serialize(
        self, value: Mapping[int, float], attr: Optional[str], obj: Any, **kwargs
    ):
        raise NotImplementedError

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ):
        if not isinstance(value, str):
            raise fields.ValidationError("Expexted a string")

        parts = config_timedelta_format.match(value)
        if not parts:
            raise fields.ValidationError("Invalid timedelta format")
        groups = parts.groupdict()
        time_params = {}
        for name, param in groups.items():
            if param:
                time_params[name] = int(param)
        return timedelta(**time_params)


CONFIG_SCHEMA = Schema.from_dict(
    name="ConfigSchema",
    fields={
        "service": fields.Nested(
            Schema.from_dict(
                {
                    "url": fields.Str(
                        required=True,
                        metadata={"description": "Base HTTP frontend URL"},
                    ),
                    "name": fields.Str(
                        required=True,
                        metadata={
                            "description": "Abrechnung instance name, will be shown e.g. in the email subjects"
                        },
                    ),
                    "api_url": fields.Str(
                        required=True, metadata={"description": "Base HTTP API URL"}
                    ),
                }
            ),
            metadata={
                "description": "settings related to the general abrechnung service"
            },
        ),
        "demo": fields.Nested(
            Schema.from_dict(
                {
                    "enabled": fields.Bool(required=False, load_default=False),
                    "wipe_interval": TimedeltaField(
                        load_default=timedelta(hours=1),
                        metadata={
                            "description": "interval after which registered users and their groups will be deleted"
                        },
                    ),
                }
            ),
            required=False,
            load_default={"enabled": False, "wipe_interval": timedelta(hours=1)},
            metadata={"description": "settings related to the demo mode"},
        ),
        "database": fields.Nested(
            Schema.from_dict(
                {
                    "host": fields.Str(required=False, load_default=None),
                    "port": fields.Int(required=False, load_default=5432),
                    "user": fields.Str(required=False, load_default=None),
                    "dbname": fields.Str(required=True),
                    "password": fields.Str(required=False, load_default=None),
                }
            )
        ),
        "api": fields.Nested(
            Schema.from_dict(
                {
                    "secret_key": fields.Str(required=True),
                    "host": fields.Str(required=True),
                    "port": fields.Int(required=True),
                    "id": fields.Str(
                        metadata={
                            "description": "id of the abrechnung api worker instance in case of multiple running"
                        },
                        required=False,
                        load_default="default",
                    ),
                    "max_uploadable_file_size": fields.Int(
                        load_default=1024,
                        required=False,
                        metadata={"description": "max file size for uploads in KB"},
                    ),  # in KB
                    "enable_cors": fields.Bool(required=False, load_default=True),
                }
            )
        ),
        "registration": fields.Nested(
            Schema.from_dict(
                {
                    "enabled": fields.Bool(
                        required=False,
                        load_default=False,
                        metadata={"description": "allow anyone to register"},
                    ),
                    "allow_guest_users": fields.Bool(
                        required=False,
                        load_default=False,
                        metadata={
                            "description": "allow guest users to register with a valid group invite"
                        },
                    ),
                    "valid_email_domains": fields.List(
                        fields.Str(),
                        required=False,
                        load_default=None,
                        metadata={
                            "description": "restrict the registration to only allow emails from specific domains"
                        },
                    ),
                }
            )
        ),
        "email": fields.Nested(
            Schema.from_dict(
                {
                    "address": fields.Str(required=True),
                    "host": fields.Str(required=True),
                    "port": fields.Int(required=True),
                    "mode": fields.Str(
                        required=False,
                        metadata={
                            "description": "email mode, one of local, smtp-ssl, smtp-starttls, smtp"
                        },
                        load_default="smtp",
                    ),
                    "auth": fields.Nested(
                        Schema.from_dict(
                            {"username": fields.Str(), "password": fields.Str()},
                            name="MailAuthSchema",
                        ),
                        required=False,
                    ),
                }
            )
        ),
    },
)


class Config:
    def __init__(self, cfg: dict):
        self._cfg = cfg

    def get(self, key, default=None):
        return self._cfg.get(key, default)

    def __getitem__(self, key):
        return self._cfg.__getitem__(key)

    @classmethod
    def from_dict(cls, cfg: dict) -> "Config":
        cfg = CONFIG_SCHEMA().load(cfg)
        return cls(cfg)

    @classmethod
    def from_file(cls, file_path: Path) -> "Config":
        with file_path.open("r") as f:
            cfg = yaml.safe_load(f)
            cfg = CONFIG_SCHEMA().load(cfg)
            return cls(cfg)



import os
from datetime import datetime

from abrechnung.application.accounts import AccountService
from abrechnung.application.groups import GroupService
from abrechnung.application.transactions import TransactionService
from tests import AsyncTestCase


class TransactionLogicTest(AsyncTestCase):
    async def setUpAsync(self) -> None:
        await super().setUpAsync()
        self.group_service = GroupService(self.db_pool, config=self.test_config)
        self.account_service = AccountService(self.db_pool, config=self.test_config)
        self.transaction_service = TransactionService(
            self.db_pool, config=self.test_config
        )

        self.user, _ = await self._create_test_user("test", "test@test.test")
        self.group_id = await self.group_service.create_group(
            user=self.user,
            name="test group",
            description="",
            currency_symbol="€",
            terms="",
            add_user_account_on_join=False,
        )

    async def _create_accounts(
        self,
        group_id: int,
        n_accounts: int,
        account_type: str = "personal",
    ) -> list[int]:
        account_ids = []
        for i in range(n_accounts):
            acc_id = await self.account_service.create_account(
                user=self.user,
                group_id=group_id,
                type=account_type,
                name=f"account{i}",
                description="",
            )
            account_ids.append(acc_id)

        return account_ids

    async def test_basic_clearing_account_workflow(self):
        basic_account_id1, basic_account_id2 = await self._create_accounts(
            self.group_id, 2
        )

        # check that we can create a simple clearing account
        account_id = await self.account_service.create_account(
            user=self.user,
            group_id=self.group_id,
            type="clearing",
            name="Clearing",
            description="Foobar",
            clearing_shares={basic_account_id1: 1.0, basic_account_id2: 2.0},
        )

        account = await self.account_service.get_account(
            user=self.user, account_id=account_id
        )
        self.assertEqual(account_id, account.id)
        self.assertFalse(account.is_wip)
        self.assertIsNotNone(account.committed_details)
        self.assertIsNone(account.pending_details)
        self.assertEqual(
            2.0, account.committed_details.clearing_shares[basic_account_id2]
        )
        self.assertEqual(
            1.0, account.committed_details.clearing_shares[basic_account_id1]
        )

        await self.account_service.update_account(
            user=self.user,
            account_id=account_id,
            name="Clearing",
            description="Foobar",
            clearing_shares={basic_account_id1: 1.0},
        )
        account = await self.account_service.get_account(
            user=self.user, account_id=account_id
        )
        self.assertIsNotNone(account.committed_details)
        self.assertIsNone(account.pending_details)
        self.assertTrue(
            basic_account_id2 not in account.committed_details.clearing_shares
        )
        self.assertFalse(account.is_wip)

    async def test_no_circular_clearing_accounts(self):
        account1_id, account2_id = await self._create_accounts(
            self.group_id, 2, account_type="clearing"
        )

        # we need to commit one account first other
        await self.account_service.update_account(
            user=self.user,
            account_id=account2_id,
            name="account2",
            description="",
            clearing_shares={account1_id: 1.0},
        )

        with self.assertRaises(Exception) as ctx:
            await self.account_service.update_account(
                user=self.user,
                account_id=account1_id,
                name="account1",
                description="",
                clearing_shares={account2_id: 1.0},
            )
        self.assertTrue(
            "this change would result in a cyclic dependency between clearing accounts"
            in str(ctx.exception)
        )

        # check that we cannot have an account reference itself
        with self.assertRaises(Exception) as ctx:
            await self.account_service.update_account(
                user=self.user,
                account_id=account1_id,
                name="account1",
                description="",
                clearing_shares={account1_id: 1.0},
            )

    async def test_file_upload(self):
        account1_id, account2_id = await self._create_accounts(self.group_id, 2)
        transaction_id = await self.transaction_service.create_transaction(
            user=self.user,
            group_id=self.group_id,
            type="purchase",
            description="foo",
            billed_at=datetime.now().date(),
            currency_symbol="€",
            currency_conversion_rate=1.0,
            value=33,
            debitor_shares={account1_id: 1.0},
            creditor_shares={account2_id: 1.0},
        )
        with open(
            os.path.join(os.path.dirname(__file__), "assets", "test_image.jpg"), "rb"
        ) as test_image:
            content = test_image.read()
            file_size = len(content)
            file_id = await self.transaction_service.upload_file(
                user=self.user,
                transaction_id=transaction_id,
                filename="test file",
                mime_type="image/jpeg",
                content=content,
            )
        transaction = await self.transaction_service.get_transaction(
            user=self.user, transaction_id=transaction_id
        )
        self.assertIsNotNone(transaction.pending_files)
        self.assertIsNone(transaction.committed_files)
        self.assertEqual(1, len(transaction.pending_files))
        self.assertEqual(file_id, transaction.pending_files[0].id)
        await self.transaction_service.commit_transaction(
            user=self.user, transaction_id=transaction_id
        )

        transaction = await self.transaction_service.get_transaction(
            user=self.user, transaction_id=transaction_id
        )
        self.assertIsNone(transaction.pending_files)
        self.assertIsNotNone(transaction.committed_files)
        self.assertEqual(1, len(transaction.committed_files))
        self.assertEqual(file_id, transaction.committed_files[0].id)
        self.assertIsNotNone(transaction.committed_files[0].blob_id)

        (mime_type, _,) = await self.transaction_service.read_file_contents(
            user=self.user,
            file_id=file_id,
            blob_id=transaction.committed_files[0].blob_id,
        )
        self.assertEqual("image/jpeg", mime_type)
        self.assertEqual(file_size, len(content))

        # now delete it
        await self.transaction_service.delete_file(user=self.user, file_id=file_id)
        transaction = await self.transaction_service.get_transaction(
            user=self.user, transaction_id=transaction_id
        )
        self.assertIsNotNone(transaction.pending_files)
        self.assertIsNotNone(transaction.committed_files)
        self.assertEqual(1, len(transaction.pending_files))
        self.assertEqual(file_id, transaction.pending_files[0].id)
        self.assertTrue(transaction.pending_files[0].deleted)
        await self.transaction_service.commit_transaction(
            user=self.user, transaction_id=transaction_id
        )
        transaction = await self.transaction_service.get_transaction(
            user=self.user, transaction_id=transaction_id
        )
        self.assertIsNone(transaction.pending_files)
        self.assertIsNotNone(transaction.committed_files)
        self.assertEqual(1, len(transaction.committed_files))
        self.assertEqual(file_id, transaction.committed_files[0].id)
        self.assertIsNone(transaction.committed_files[0].blob_id)
        self.assertTrue(transaction.committed_files[0].deleted)
